#!/usr/bin/env python

import numpy as np
from cca import FloatRange, LARGE_STEPS, OnlyFloatAbove, OnlyIntegerRange, OnlyIntegerAbove, OnlyIntegerBelow
import sys
      

class Prop(object):
    def __init__(self, name, **kwargs):
        self.name = "_" + name
        self.changed = self.name + "_changed"

    def __get__(self, obj, type=None):
        return getattr(obj, self.name)

    def __set__(self, obj, val):
        v = self.constrain(self.parse(val))
        old_v = None
        if hasattr(obj, self.name):
            old_v = getattr(obj, self.name)
        if v != old_v:
            print "Changed from %s to %s." % (repr(old_v), repr(v))
            setattr(obj, self.changed, True)
            setattr(obj, self.name, v)

    def parse(self, val):
        return int(val)

    def constrain(self, val):
        return val

    def test(self, obj):
        print "Hei", obj.__class__.__name__, getattr(obj, self.name)


class PropA(Prop):
    def constrain(self, val):
        if val < 0:
            val = 0
        return super(PropA, self).constrain(val)


class PropB(Prop):
    def constrain(self, val):
        if val > 700:
            val = 700
        return super(PropB, self).constrain(val)


class PropC(PropA, PropB):
    pass

class PropD(PropA, PropB):
    def parse(self, val):
        return float(val)

class ClassWithParams(type):  # Metaclass
    def __init__(cls, name, bases, dct):
        super(ClassWithParams, cls).__init__(name, bases, dct)
        d = {}
        for c in cls.__mro__:
            for (k, v) in c.__dict__.items():
                if issubclass(v.__class__, Prop):
                    d[k] = v
        cls.param = d
        
class A(object):
    __metaclass__ = ClassWithParams
    p = PropC("abc")

class B(A):
    q = PropD("dsf")
    v = PropD("asd")

a = B()
a.p = a.q = a.v = 32.12
a.v = 2.2
print a.p, a.q, a.v
a.__class__.param["p"].test(a)

if 0:
    class Parent:
        def __init__(self):
            self.param = {}
            r = OnlyIntegerRange(self, start=1, stop=4, default=3)
            self.param["abcd"] = r
            print r.value
            r.prod(-1, LARGE_STEPS)
            r.prod(-1, LARGE_STEPS)
            print r.value
            r.prod(1, LARGE_STEPS)
            r.prod(1, LARGE_STEPS)
            print r.value
            
            
    p = Parent()
            
