#!/usr/bin/env python

import hoca
from interface import CA_wrapper, CA_set_wrapper, Display, start
from random import randint
import sys

DEBUG=True

class OrganActivityRule:
    def __call__(self, lattice, x):
        cells = lattice.down.down(x)
        sublat = lattice.down.below
        substates = lattice.down_states(x)
        return (sum(substates),)

class ActivityTMap:
    def __init__(self, target=3, tolerance=1):
        self.target = 3
        self.tolerance = 1
    
    def __call__(self, lattice, x):
        avg = lattice.up_states(x)[0]
        if abs(avg-self.target) > self.tolerance:
            #return (0, (lattice.rule[x]+1)%len(lattice.twoRule))
            return (0, (lattice.rule[x]+1)%len(lattice.twoRule))
        return (0, lattice.rule[x])


def my_example(**params):
    if DEBUG:
        print params
    pres = {}
    for (k, v) in params.iteritems():
        val = v
        if type(v) == type([]):
            if type(v[0]) == type([]):
                val = "-".join(["".join(["%d" % i for i in x]) for x in v])
            else:
                val ="".join(["%s" % x for x in v])
        pres[k] = val
    pats = CA_set_wrapper("-".join(["%s%s" % x for x in pres.iteritems()]))

    w = params.get("width", 100)
    mod = params.get("mod", 10)
    w = w - w % mod
    max_state = params.get("max_state", 1)
    radius = params.get("radius", 2)
    max_sum = (2*radius + 1)*max_state

    l2 = hoca.Lattice(((),), (OrganActivityRule(),), hoca.NullTMap(), 
                      mod, w/mod, ("uint8",))
    
    two_nbh = ( hoca.radiusNbh(radius), )
    two_rule = []
    maps = params.get("maps", None)
    if maps == None:
        for i in range(mod):
            arr = randarr(max_sum+1,0,max_state)
            if DEBUG:
                print arr
            two_rule.append(hoca.SumMapRule(arr))
    else:
        two_rule = [hoca.SumMapRule(x) for x in maps]
    tmap = ActivityTMap(params.get("target", 2), 
                        params.get("tolerance", 2))
    l1 = hoca.Lattice(two_nbh, two_rule, tmap, max_state, w)
    l1.random(rule=False)

    M = hoca.ModMap(mod)
    hoca.connect(l1, l2, M)

    state = hoca.LatticeStack((l1,l2))

    subtext = "" #" ".join(["%s:%s" % x for x in pres.iteritems()])
    pats.append(CA_wrapper(state, "HOCA", subtext))

    pats[-1].display.append(Display("state", 0, 0, "255*(dmax-data)/dmax", True))
    pats[-1].display.append(Display("rule", 0, None, "255*(dmax-data)/dmax", True))
    pats[-1].display.append(Display("state", 1, 0, "hoca.enlarge(255*(dmax-data)/dmax,%d)" % mod, True))

    return pats

def randarr(n, a, b):
    return [randint(a, b) for x in range(n)]

def random_list():
    while 1:
        maxmod = 5
        mod = randint(2, maxmod)
        radius = randint(1,2)
        max_state = 1
        target = randint(0, mod-1)
        tolerance = randint(0, 2)
        yield my_example(**{"radius": radius,
                            "mod": mod,
                            "max_state": max_state,
                            "target": target,
                            "tolerance": tolerance})

def my_list():
    for i in range(0, 5):
        yield my_example(**{"radius": 2,
                            "mod": 5,
                            "max_state": 1,
                            "target": i,
                            "tolerance": 0,
                            "maps": [[0, 0, 0, 0, 1, 1],
                                     [0, 1, 1, 1, 0, 0],
                                     [1, 0, 0, 1, 1, 0],
                                     [1, 0, 0, 1, 0, 0],
                                     [1, 1, 0, 1, 1, 1]
                                     ]})


args = {"iterator": my_list(),
        "height": 400}

start(**args)
