#!/usr/bin/env python

#
# CA Explorer program for exploring cellular automata
# Copyright (C) 2009 Lauritz Thaulow
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

START_MESSAGE = """
  CA Explorer is Copyright (C) 2009 Lauritz Thaulow
  This program comes with ABSOLUTELY NO WARRANTY; for details,
  read the "About" dialog.

  This is free software, and you are welcome to redistribute it
  under certain conditions; see the "About" dialog for the license.
"""

VERSION = "0.1"

import wx
import os
import sys
from math import log10, ceil
from time import time, sleep
from types import TupleType, IntType, FloatType
import numpy as np
import hoca
from random import randint
pjoin = os.path.join

ID_SAVE = 102
ID_ABOUT = 101
ID_EXIT = 110



class CA_wrapper:
    def __init__(self, state, ca_type="CA", caption="Caption"):
        self.state = state
        self.caption = caption
        self.display = []


class CA_set_wrapper(list):
    def __init__(self, name="Filename"):
        super(list, self).__init__()
        self.name = name


class Display:
    def __init__(self, name, level, layer, eval_str, calculate_bounds=False):
        self.name = name
        self.level = level
        self.layer = layer
        self.eval = eval_str
        self.calculate_bounds = calculate_bounds

    def get_bounds(self, data):
        dmax=np.max(data)
        return dmax

    def get_data(self, wrapper):
        lat = wrapper.state[self.level]
        data = np.array(getattr(lat, self.name), 
                        dtype="uint32")
        if self.layer != None:
            data = data[self.layer]
            dmax = lat.max[self.layer]
        if self.name == "rule":
            dmax = len(lat.twoRule)
        if self.name == "nbh":
            dmax = len(lat.twoNbh)
        return eval(self.eval)



def example1(w=200, ca=3, cb=3):
    two_nbh = ( hoca.radiusNbh(2), )
    a = hoca.SumMapRule((0,0,1,1,0,0))
    b = hoca.SumMapRule((0,1,0,0,0,0))
    two_rule = (a, b)
    tmap = hoca.TotExTMap(ca, cb)
    l = hoca.Lattice(two_nbh, two_rule, tmap, 1, w)
    state = hoca.LatticeStack(l)
    display = (("state", 0, 0, "255*(1-data)"), 
               ("rule", 0, None, "255*data"))
    return ((state, display),)

def example2(w=200, k=2, r=(2,4), s=(2,6), map=(0,1,0,1,0,1,0,1,0,1)):
    two_nbh = [hoca.radiusNbh(x) for x in range(r[0],r[1]+1)]
    two_rule = (hoca.SumMapRule(map),)
    tmap = hoca.EquiTotTMap(*(s + r))
    l = hoca.Lattice(two_nbh, two_rule, tmap, 1, w)
    state = hoca.LatticeStack(l)
    display = (("state", 0, 0, "%d*(%d-data)" % (255/(k-1), (k-1))), 
               ("nbh", 0, None, "%d*data" % (255/(r[1]-r[0]))))
    return ((state, display),)

def example3(ra, rb):
    ret = CA_set_wrapper("example3")
    M = hoca.ModMap(19)
    lvl1_2nbh = (hoca.radiusNbh(2),)
    lvl2_2nbh = (hoca.radiusNbh(1),)
    lvl1_2rule = [hoca.SumMapRule(x) for x in (ra, rb)]

    class lvl2_rule:
        def __init__(self, map):
            self.map = [int(x) for x in map]
        def __call__(self, lattice, x):
            parity = sum(lattice.down_states(x)) % 2
            s = lattice.sum(x)
            return (self.map[parity*4+s],)
    lvl2_2rule = (lvl2_rule("00101110"),)

    def tmap(lattice, x):
        return (0, lattice.up_states(x)[0])

    l1 = hoca.Lattice(lvl1_2nbh, lvl1_2rule, tmap, 1, 589)
    l2 = hoca.Lattice(lvl2_2nbh, lvl2_2rule, hoca.NullTMap(), 1, 31)
    hoca.connect(l1, l2, M)
    state = hoca.LatticeStack((l1, l2))
    state.random()
    ca_wrapper = CA_wrapper(state, "HOCA", "asdflkj")
    ca_wrapper.display.append(Display("state", 0, 0, "255*(1-data)"))
    ca_wrapper.display.append(Display("rule", 0, None, "255*data"))
    ca_wrapper.display.append(Display("state", 1, 0, "hoca.enlarge(255*data, 19)"))
    ret.append(ca_wrapper)
                              
    return ret

def my_example1(w=300, r=None, c=None):
    pats = []
    if r == None:
        r = (randint(0,255), randint(0,255))
    if c == None:
        c = (randint(0,3), randint(0,3))
    print r, c
    two_nbh = ( hoca.radiusNbh(1), )
    two_rule = []
    for rule in r:
        f = hoca.WolframRule(rule)
        l = hoca.Lattice(two_nbh, (f,), hoca.NullTMap(), 1, w)
        l.random()
        ls = hoca.LatticeStack(l)
        ls.identifier = "Rule %d" % rule
        #pats.append([ls, (("state", 0, 0, "255*(1-data)"),)])
        two_rule.append(f)
    tmap = hoca.TotExTMap(*c)
    l = hoca.Lattice(two_nbh, two_rule, tmap, 1, w)
    l.state[0][:] = 0
    ini = np.random.randint(0,2,(10,))
    #ini = np.fromiter((0,0,1,0,0,1,0), dtype="uint8")
    l.state[0][w/2:w/2+10] = ini
    #l.random(rule=False)
    state = hoca.LatticeStack(l)
    display = (("state", 0, 0, "255*(1-data)"), 
               )
#               ("rule", 0, None, "255*(1-data)"))
    state.identifier = {#"ca":c[0], "cb":c[1], "fa":r[0], "fb":r[1],
                       "ini": "".join(["%d" % x for x in ini.tolist()])}
    pats.append([state, display])
    return pats

def my_example2():
    M = hoca.ModMap(8)
    lvl1_2nbh = ( hoca.radiusNbh(1), )
    lvl1_2rule = [hoca.WolframRule(x) for x in range(256)]
    def tmap(lattice, x):
        s_up = lattice.up_states(x)[0]
        if s_up == 0:
            return (0, lattice.rule[x] + 1)
        if s_up == 255:
            return (0, lattice.rule[x] - 1)
        return (0, lattice.rule[x])
    l1 = hoca.Lattice(lvl1_2nbh, lvl1_2rule, tmap, 1, 256)

    lvl2_2nbh = ( (0,), )
    def lvl2_rule(lattice, x):
        s = hoca.statify(lattice.down_states(x))
        return s
    l2 = hoca.Lattice(lvl2_2nbh, (lvl2_rule,), hoca.NullTMap(), 255, 32)

    hoca.connect(l1,l2,M)
    state = hoca.LatticeStack((l1,l2))
    display = (("state", 0, 0, "255*(1-data)"), 
               ("rule", 0, None, "data"),
               ("state", 1, 0, "hoca.enlarge(data, 8)"))
    
    return ((state, display),)

def my_example3(w=200):
    pats = []
    two_nbh = ( hoca.radiusNbh(4), )
    two_rule = ( hoca.SumMapRule(273), )
    l = hoca.Lattice(two_nbh, two_rule, hoca.NullTMap(), 1, w)
    l.state[0][:] = 0
    l.state[0][w/2] = 1
    #l.random(rule=False)
    state = hoca.LatticeStack(l)
    display = (("state", 0, 0, "255*(1-data)"),)
    state.identifier = {"r":4, "s":2, "code":273}
    pats.append([state, display])

    from equivalent_r1_ca import get_1r_equivalent_map
    two_nbh = ( hoca.radiusNbh(1), )
    two_rule = ( hoca.WolframMapRule(get_1r_equivalent_map(273)), )
    l = hoca.Lattice(two_nbh, two_rule, hoca.NullTMap(), 15, w/4)
    l.state[0][:] = 0
    l.state[0][w/(4*2)] = 1
    #l.random(rule=False)
    state = hoca.LatticeStack(l)
    display = (("state", 0, 0, "16*(15-hoca.enlarge(data,4))"),)
    state.identifier = {"r":1, "s":16, "code_equiv":273}
    pats.append([state, display])
    
    return pats


def examples():
    yield my_example3(200)
    return
    while 1:
        yield my_example1(200, (241,94), (1,2))
    #return
    for (r, c) in (
        ((86,241),(2,2)),
        ((114,148),(3,3)),
        ((1,36),(1,1)),
        ((28,145),(1,1)),
        ((33,38),(3,1)),
        ((46,123),(1,3)),
        ((55,250),(2,2)),
        ((88,182),(3,1)),
        ((89,131),(1,0)),
        ((98,75),(2,3)),
        ((114,148),(3,3)),
        ((122,69),(3,1)),
        ((132,85),(1,1)),
        ((150,126),(0,2)),
        ((159,2),(3,0)),
        ((178,164),(2,1)),
        ((210,210),(2,2)),
        ((212,51),(2,2)),
        ((234,24),(3,1)),
        ((114,148),(3,3)),
        ((241,94),(1,2)),
        ((241,103),(3,3)),
        #((255,28),(3,2)),
        ):
        yield my_example1(200, r, c)

def id_to_text(x):
    if type(x) == type(""):
        return x
    l = []
    keys = x.keys()
    keys.sort()
    for k in keys:
        l.append("%s: %s" % (k, x[k]))
    return ", ".join(l)

def id_to_filename(x):
    if type(x) == type(""):
        return x
    l = []
    keys = x.keys()
    keys.sort()
    for k in keys:
        l.append("%s-%s" % (k, x[k]))
    return "_".join(l)


class PatternPanel(wx.Panel):
    def __init__(self, parent, **kwargs):
        wx.Panel.__init__(self, parent, wx.ID_ANY, size=(440,680), style=wx.BG_STYLE_CUSTOM|wx.TAB_TRAVERSAL)
        self.kwargs = kwargs
        self.parent = parent
        self.SetBackgroundColour('white')
        self.Bind(wx.EVT_PAINT, self.OnPaint)
        self.do_draw = True
        self.Bind(wx.EVT_CHAR, self.OnChar)
        self.timer = wx.Timer()
        self.timer.SetOwner(self)
        self.Bind(wx.EVT_TIMER, self.OnTimer) 

        self.h = kwargs.get("height", 100)
        self.iterator = kwargs.get("iterator", iter([]))
        self.reset_cas()


    def reset_cas(self):
        self.cas = self.iterator.next()

        self.margin = 8
        self.sep = 10
        self.bot = 50

        self.w = self.cas[0].state[0].width
        self.w_sep = self.w + self.sep
        self.stripes = sum([len(x.display) for x in self.cas])
        self.w_tot = 2*self.margin + self.stripes*self.w + \
            (self.stripes-1)*self.sep
        self.h_tot = self.h + 2*self.margin + self.bot

        self.img = np.zeros((self.h_tot, self.w_tot, 3), dtype="uint8")
        self.img[...] = 255
        self.im = wx.EmptyImage(self.w_tot, self.h_tot)

                                
    def OnTimer(self, e):
        self.PaintPattern()

    def PaintPattern(self):
        if not self.do_draw:
            return

        for i in range(self.h):
            col_no = 0
            for ca_wrapper in self.cas:
                for disp in ca_wrapper.display:
                    data = disp.get_data(ca_wrapper)
                    start = self.margin + col_no*self.w_sep
                    stop = start + self.w
                    for c in (0,1,2):
                        self.img[i+self.margin,start:stop,c] = data
                    col_no += 1
                ca_wrapper.state = ca_wrapper.state.next() 
        self.im.SetData(self.img.tostring())
        bm = self.im.ConvertToBitmap()
        self.memdc = wx.MemoryDC(bm)
        self.memdc.SetBrush(wx.TRANSPARENT_BRUSH)
        
        col_no = 0
        for ca_wrapper in self.cas:
            curr_h = self.margin + self.h + 2
            num_in_box = len(ca_wrapper.display)
            for disp in ca_wrapper.display:
                text = disp.name.capitalize()
                (width, height) = self.memdc.GetTextExtent(text)
                self.memdc.DrawText(text, self.margin + col_no*self.w_sep + 
                                    self.w/2 - width/2, curr_h)
                col_no += 1
            text = ca_wrapper.caption
            (tw, th) = self.memdc.GetTextExtent(text)
            curr_h += th + 2
            left_edge = self.margin + (col_no-num_in_box)*self.w_sep 
            width = num_in_box*self.w_sep - self.sep
            middle = left_edge + width/2
            self.memdc.DrawText(text, middle - tw/2, curr_h)
            self.memdc.DrawRectangle(left_edge-4,self.margin-4,
                                     width+8,self.h_tot-self.margin-
                                     2*self.margin+8)

        dc = wx.BufferedDC(None, self.buf)
        dc.Blit(0, 0, self.w_tot, self.h_tot, self.memdc, 0, 0)
        del dc
        self.OnPaint()
        tosave = bm.ConvertToImage()
        tosave.SetOption("quality", "85")
        tosave.SaveFile(pjoin("data", self.cas.name + ".png"),
                        wx.BITMAP_TYPE_PNG)
        try:
            self.reset_cas()
        except StopIteration:
            self.do_draw = False
            return
        self.timer.Start(1000)


    def updateSize(self):
        (self.width, self.height) = self.GetClientSizeTuple()
        return (self.width, self.height)

    def OnResize(self, evt):
        width, height = self.updateSize()
        self.buf = wx.EmptyBitmap(width, height)
        dc = wx.BufferedDC(None, self.buf)
        dc.Clear()
        del dc
        self.PaintPattern()
        #self.Reset()

    def OnPaint(self, evt=None):
        if not hasattr(self, "buf"):
            self.OnResize(None)
        wx.BufferedPaintDC(self, self.buf)

    def OnChar(self, e):
        kc = e.GetKeyCode()
        print kc
        if kc == 32:
            self.h = 500
        if kc in (65, 97):  # Aa
            (self.initial, self.display) = \
                my_example1()
        if kc in (66, 98):  # Bb
            (self.initial, self.display) = \
                my_example2()
        self.PaintPattern()
        

class AboutDialog(wx.Dialog):
    def __init__(self):
        wx.Dialog.__init__(self, None, wx.ID_ANY, "About CA Explorer", 
                           size=(640, 480), style=wx.DEFAULT_DIALOG_STYLE)
        sizer = wx.BoxSizer(wx.VERTICAL)
        text = "\nCA Explorer %s is (C) 2009 Lauritz Thaulow\n" % VERSION
        sizer.Add(wx.StaticText(self, wx.ID_ANY, text))
        path = os.path.dirname(sys.argv[0])
        license = open(os.path.join(path, "LICENSE.txt"), "r").read()
        content = wx.TextCtrl(self, wx.ID_ANY, license, (0,0), (640, -1),
                              style=wx.TE_READONLY|wx.TE_MULTILINE)
        content.SetFont(wx.Font(10, wx.FONTFAMILY_TELETYPE, 
                                wx.FONTSTYLE_NORMAL, wx.FONTWEIGHT_NORMAL))
        sizer.Add(content, 1, wx.EXPAND)
        ok = wx.Button(self, wx.ID_ANY, "OK")
        self.Bind(wx.EVT_BUTTON, self.OnOK)
        sizer.Add(ok, 0, wx.ALIGN_CENTER)
        self.SetSizer(sizer)
        content.SetInsertionPoint(0)
        
    def OnOK(self, evt):
        self.Destroy()


class RectDialog(wx.Dialog):
    def __init__(self, vars):
        wx.Dialog.__init__(self, None, wx.ID_ANY, "Specify dimensions", 
                           size=(640, 480), 
                           style=wx.DEFAULT_DIALOG_STYLE)
        self.vars = vars
        self.inp = {}
        sizer = wx.BoxSizer(wx.VERTICAL)
        for v in vars:
            sizer2 = wx.BoxSizer(wx.HORIZONTAL)            
            sizer2.Add(wx.StaticText(self, wx.ID_ANY, "%s: " % v))
            self.inp[v] = wx.TextCtrl(self, wx.ID_ANY, "%d" % self.vars[v])
            sizer2.Add(self.inp[v])
            sizer.Add(sizer2)
        ok = wx.Button(self, wx.ID_ANY, "OK")
        self.Bind(wx.EVT_BUTTON, self.OnOK)
        sizer.Add(ok, 0, wx.ALIGN_CENTER)
        self.SetSizer(sizer)
        
    def OnOK(self, evt):
        for v in self.vars:
            self.vars[v] = int(self.inp[v].GetValue())
        self.Destroy()


class MainWindow(wx.Frame):
    def __init__(self, parent, id, title, **kwargs):
        wx.Frame.__init__(self, parent, id, title, size=(1000, 680))
        self.args = {}
        self.changed = False
        self.curr_clss = None
        filemenu= wx.Menu()
        filemenu.Append(ID_ABOUT, "&About"," Information about this program")
        filemenu.AppendSeparator()
        filemenu.Append(ID_EXIT,"&Quit"," Exit the program.")
        menuBar = wx.MenuBar()
        menuBar.Append(filemenu,"&File")
        self.SetMenuBar(menuBar)

        self.pp = PatternPanel(self, **kwargs)
        self.sizer = wx.BoxSizer(wx.VERTICAL)
        self.sizer.Add(self.pp, 1, wx.EXPAND)
        self.SetAutoLayout(True)
        self.SetSizer(self.sizer)
        self.Layout()
        #self.sizer.Fit(self)

        self.SetMinSize(wx.Size(10,10))
        self.CreateStatusBar()
                  
        wx.EVT_MENU(self, ID_ABOUT, self.OnAbout)
        wx.EVT_MENU(self, ID_EXIT, self.OnExit)
        #self.Bind(wx.EVT_IDLE, self.OnIdle)

        self.Show(True)
        self.pp.Bind(wx.EVT_SIZE, self.pp.OnResize)
        self.SetFocus()
        #self.pp.OnResize(None)
        #self.pp.PaintPattern(self.args)
        wx.EVT_CLOSE(self, self.OnClose)

    def OnAbout(self, e):
        d = AboutDialog()
        d.ShowModal()
        d.Destroy()

    def OnExit(self, e):
        self.Close(True)  # Close the frame.

    def OnClose(self, e):
        print "exiting...."
        self.pp.do_draw = False
        self.Destroy()



def start(**kwargs):
    print START_MESSAGE
    import sys,os
    app = wx.PySimpleApp()
    frame = MainWindow(None, wx.ID_ANY, "HOCACA Explorer", **kwargs)
    app.MainLoop()


if __name__ == '__main__':
    #import cProfile
    #cProfile.run('start()', 'profile.txt')
    def ite():
        yield example3("010111", "001010")
        yield example3("001100", "010000")
    start(iterator=ite(), height=300)
    


