
#include "rules.h"
#include <string.h>
#include <iostream>
#include <iomanip>
#include <stdexcept>
#include <cstring>


State::State(const Coord &dims, const Nbhs_ptr n, const Rules r, const unsigned int params_):
        width(dims.first), height(dims.second), params(params_),
        x_stride(params_), y_stride(x_stride*dims.first),
        arr_size(y_stride*dims.second), nbhs(n), rules(r),
        arr(new unsigned int [arr_size]) {
}

void State::apply(State &next) {
    TFunctor *rule;
    //min_c = this->max_c; max_c = 0;
    //next.clear(ChangesIdx);
    //unsigned long int changes = 0;
    for (unsigned int y = 0; y < height; y++) {
        for (unsigned int x = 0; x < width; x++) {
            //rule = static_cast<TFunctor*>(rules.at(arr[t + 2*width + x]));
            rule = rules.at(arr[x*x_stride + y*y_stride + RuleIdx]);
            rule->apply(*this, next, x, y);
            //changes += arr[x*x_stride + y*y_stride + ChangesIdx];
        }
    }
    //std::cout << changes << std::endl;
}

void State::sums(unsigned int x, unsigned int y, unsigned int max_state, unsigned int &inf, unsigned int &ill,
                 unsigned int &sum, unsigned int &state, unsigned long int &arr_index) {
    arr_index = x*x_stride + y*y_stride;
    inf = 0; ill = 0; sum = 0;

    int nbh_i = arr[arr_index + NbhIdx];   // neighbourhood index is the first parameter
    Coords::iterator ni, ne;          // [n]eighbourhood [i]terator and [n]eighbourhood [e]nd.
    ne = (*nbhs)[nbh_i].end();
    long int comp_x = 0, comp_y = 0;
    for (ni = (*nbhs)[nbh_i].begin(); ni != ne; ni++) {
//        if (ni->first == 0 && ni->second == 0) {
//            continue;
//        }
        comp_x = ((x + ni->first)*x_stride + y_stride) % y_stride;
        comp_y = ((y + ni->second)*y_stride + arr_size) % arr_size;
        state = arr[comp_x + comp_y];
        sum += state;
        if (state >= max_state) {
            ill += 1;
        }
        if (state != 0) {
            inf += 1;
        }
    }
    state = arr[arr_index];
}

unsigned int & State::mod_idx(unsigned int x, unsigned int y, unsigned int mod, Parameter p) {
    unsigned long int arr_index = (x - x%mod)*x_stride + (y - y%mod)*y_stride + p;
    return arr[arr_index];
}

void State::print(void) {
    for (unsigned int y = 0; y < height; y++) {
        for (unsigned int x = 0; x < width; x++) {
            std::cout << std::setw(4) << static_cast<int>(arr[x*x_stride + y*y_stride]) << " ";
        }
        std::cout << std::endl;
    }
    std::cout << std::endl;
}

State * State::clone(void) {
    Coord c = Coord(width, height);
    State * st = new State(c, nbhs, rules, params);
    std::memcpy(st->get(), arr.get(), sizeof(unsigned int)*arr_size);
    return st;
}

void State::random(unsigned int perc, unsigned int param, unsigned int min, unsigned int max) {
    unsigned int sentinel_i = param + arr_size;
    if (perc == 0) {
        for (int i = param; i != sentinel_i; i += x_stride) {
            arr[i] = 0;
        }
        return;
    }
    unsigned int var = max - min;
    for (int i = param; i != sentinel_i; i += x_stride) {
        if (rand() % 100 >= perc) {
            arr[i] = 0;
        } else {
            arr[i] = rand() % var + min;
        }
    }
}

void State::clear(int param) {
    if (param != -1) {
        random(0, param);
    } else {
        for (int i = 0; i != params; i++) {
            random(0, i);
        }
    }
}

void State::read_state_from(State &st) {
    unsigned int * b = this->begin();
    unsigned int * e = this->end();
    unsigned int * ob = st.begin();
    unsigned int * oe = st.end();
    while (b != e) {
        *b++ = *ob++;
    }
}

//unsigned int State::get_max_state() {
//    return rules.at(0)->get_n();
//}

Rule::Rule(double _k1, double _k2, unsigned int _g, unsigned int _n): k1(_k1), k2(_k2), g(_g), n(_n) {}

unsigned int Rule::apply(State &src, State &tar, unsigned int x, unsigned int y) {
    unsigned int state = 0, inf = 0, ill = 0;
    unsigned int sum = 0;
    unsigned long int arr_index = 0;
    src.sums(x, y, n, inf, ill, sum, state, arr_index);  // Sets the last 5 parameters

    unsigned int new_state;
    if (state == 0) {
        new_state = static_cast<unsigned int>(inf/k1 + ill/k2);
    } else if (state >= n) {
        new_state = 0;
    } else {
        if (inf == 0) {
            new_state = g;
        } else {
            new_state = static_cast<unsigned int>(g + sum/inf);
        }
    }
    if (new_state > n-1) {
        new_state = n;
    }

    tar.get()[arr_index + StateIdx] = new_state;

    return new_state;
}

unsigned int ZRule::apply(State &src, State &tar, unsigned int x, unsigned int y) {
    unsigned long int arr_index = x*src.x_stride + y*src.y_stride;
    unsigned int changes = src.get()[arr_index + ChangesIdx];
    if (changes > 250) {
        tar.get()[arr_index + ChangesIdx] = 0;
        tar.get()[arr_index + StateIdx] = 0;
        tar.get()[arr_index + RuleIdx] = 1;
    } else {
        if (changes > 2) {
            tar.get()[arr_index + RuleIdx] = 0;
        } else {
            tar.get()[arr_index + RuleIdx] = 1;
        }
        tar.get()[arr_index + ChangesIdx] = changes + 1;
    }
    return 0;
}


CA1D::CA1D(State_ptr &s):
        state(State_ptr(s->clone())), next(State_ptr(s->clone())),
        width(s->width), height(s->height) {

}

State_ptr CA1D::get(bool advance) {
    if (advance) {
        state->apply(*next);
        state.swap(next);
    }
    return state;
}



