 
#ifndef RULES_H
#define RULES_H

#include <vector>
#include <boost/shared_array.hpp>
#include <boost/shared_ptr.hpp>
#include <boost/ptr_container/ptr_vector.hpp>


enum Parameter { StateIdx, NbhIdx, RuleIdx, ChangesIdx };

class State;
typedef boost::shared_ptr<State> State_ptr;

typedef std::pair<int, int> Coord;
typedef std::vector<Coord> Coords;
typedef std::vector<Coords> Nbhs;

typedef boost::shared_ptr<Nbhs> Nbhs_ptr;


typedef unsigned int (*fptr)(State &src, State &tar, unsigned int x, unsigned int y);

class TFunctor
{
public:
  virtual unsigned int apply(State &src, State &tar, unsigned int x, unsigned int y)=0;        // call using function
//  virtual unsigned int get_n()=0;
};

// derived template class
template <class TClass> class TSpecificFunctor : public TFunctor
{
private:
  unsigned int (TClass::*fpt)(State &src, State &tar, unsigned int x, unsigned int y);   // pointer to member function
  TClass* pt2Object;                  // pointer to object

public:
  // constructor - takes pointer to an object and pointer to a member and stores
  // them in two private variables
  TSpecificFunctor(TClass* _pt2Object, unsigned int(TClass::*_fpt)(State &src, State &tar, unsigned int x, unsigned int y))
     { pt2Object = _pt2Object;  fpt=_fpt; }

  virtual unsigned int apply(State &src, State &tar, unsigned int x, unsigned int y)
    { return (*pt2Object.*fpt)(src, tar, x, y);}             // execute member function

//  virtual unsigned int get_n()
//    { return pt2Object->n; }
};

typedef std::vector < TFunctor* > Rules;

class Rule;

class State
{
public:
    State(const Coord &dims, const Nbhs_ptr n, const Rules r, const unsigned int params_=5);

    const unsigned int width, height, params, x_stride, y_stride, arr_size;
    //unsigned int min_c, max_c;

    unsigned int * begin(void)   { return arr.get(); }
    unsigned int * end(void)     { return arr.get() + arr_size; }
    unsigned int * get(void)     { return arr.get(); }
    unsigned int & mod_idx(unsigned int x, unsigned int y, unsigned int mod, Parameter p);

    void print(void);
    State * clone(void);
    unsigned int * scanLine(unsigned int y) { return arr.get() + y*y_stride; }

    void apply(State &next);
    void sums(unsigned int x, unsigned int y, unsigned int max_state, unsigned int &inf, unsigned int &ill,
              unsigned int &sum, unsigned int &state, unsigned long int &arr_index);
    void random(unsigned int perc=0, unsigned int param=0, unsigned int min=0, unsigned int max=1);
    void clear(int param=-1);
    void read_state_from(State &st);
    unsigned int get_max_state();

  private:
    const Nbhs_ptr nbhs;
    const Rules rules;
    boost::shared_array<unsigned int> arr;
};

class Rule
{
 public:
    Rule(double _k1, double _k2, unsigned int _g, unsigned int _n);
    double k1, k2;
    unsigned int g, n;

    unsigned int apply(State &src, State &tar, unsigned int x, unsigned int y);
};

class ZRule
{
 public:
    ZRule() {}

    unsigned int apply(State &src, State &tar, unsigned int x, unsigned int y);
};

class CA1D
{
 public:
    CA1D(State_ptr &s);
    State_ptr get(bool advance=true);
    
 private:
    boost::shared_ptr<State> state, next;

 public:
   const unsigned int width, height;
   unsigned int x, y;
   double k1, k2;
   unsigned int g, n;
};

typedef boost::shared_ptr<CA1D> CA1D_ptr;

#endif
