#!/usr/bin/env python

import sys
from pdfdraw import Make_PDF, mm, pt
from collections import defaultdict
import os
from ca import *

rules = set(range(256))


def emptystring():
    return ""

def int2bin(n, count=8):
    return "".join([str((n >> y) & 1) for y in range(count-1, -1, -1)])

def swapbits(n, i, j):
    tmp = ((n >> i) ^ (n >> j)) & 1
    return n ^ ((tmp << i) | (tmp << j))

def mirror(rule):
    return swapbits(swapbits(rule, 1, 4), 3, 6)

def invert(rule):
    r = 0
    for i in range(8):
        r = (r << 1) | ((rule & 1) ^ 1)
        rule = rule >> 1
    return r

def new_pdf_gen():
    grid = "6x10"
    pagesize = "150x220"
    count = 1
    while 1:
        yield Make_PDF("1DCA_p%d.pdf" % count, grid=grid, pagesize=pagesize,
                        caption="The 1D cellular automatas, page %d" % count,
                        placement="p", draw_box=True, width="\\textwidth")
        count += 1
        



if __name__ == "__main__":

    ratio = 35/20.0
    iter = 22

    new_pdf = new_pdf_gen()
    pdf = new_pdf.next()
    while len(rules):
        rule = rules.pop()
        similar = defaultdict(emptystring)
        similar[mirror(rule)] += "m"
        similar[invert(rule)] += "i"
        similar[mirror(invert(rule))] += "M"
        rules -= set(similar.keys())
        ruletype = len(similar[rule]) and " (%s)" % similar[rule] or ""
        label = "Rule %d%s\n" % (rule, ruletype)
        del similar[rule]
        if len(similar) > 0:
            keys = similar.keys()
            keys.sort()
            sim = ", ".join(["%d (%s)" % (k, similar[k]) for k in keys])
            label += "%s" % (sim)
        r = CA1D_Rule(rule)
        size = int(iter*ratio)
        if size % 2 != 0:
            size += 1
        for l in (CA1D_Lattice(size=size, initial="*1*"),
                  CA1D_Lattice(size=size, initial="*100001100001110010101110111111*")):
            if not pdf.translate():
                pdf.save()
                pdf = new_pdf.next()
                pdf.translate()
            pdf.draw_ca(CA1D(lattice=l, rule=r, iterations=iter),
                        (0,8), (pdf.div_width, pdf.div_height), "")
        labels = label.split("\n")
        label_space = pdf.font_size*pt*(len(labels)+0.4)
        if label == "":
            label_space = 0
        c = pdf.c
        cs = pdf.cellspacing
        c.resetTransforms()
        c.translate(cs + (pdf.x_div-1)*(pdf.div_width+cs),
                    cs + (pdf.y_divs - pdf.y_div - 1)*(pdf.div_height + cs))
        c.setFillColorRGB(0.9,0.9,1)
        w = 2*pdf.div_width + cs
        c.setLineWidth(0.1); c.rect(0,0,w,label_space,stroke=0,fill=1)
        c.setFillColorRGB(0,0,0)
        c.setFont("Helvetica", pdf.font_size)
        for (i,l) in enumerate(labels):
            c.drawCentredString(w/2, label_space-(i+1)*pdf.font_size*pt, l)
            
    pdf.save()
        
        
    
