#!/usr/bin/env python

import WatorHOCA as hoca
from WatorInterface import CA_wrapper, Display, start

def example(dim):
    (w, h) = dim
    
    # First level lattice, rules, neighbourhoods
    two_nbh = (hoca.vonNeumann(),)
    two_rule = (hoca.WatorRule(8, 20, 20),)
    l1 = hoca.Lattice(two_nbh, two_rule, hoca.NullTMap(), max_state=2, 
                      dim=(w,h), dtypes=("uint8",)*3 + ("bool",), 
                      inplace=True, on_next="self.state[3][:,:] = False; ")
    
    # Put it all together, and specify which parts to display
    ret = CA_wrapper(hoca.LatticeStack(l1), "CA")
    
    ret.display.append(Display("state", 0, 0, "127*(2-data)"))
    return [ret]


def yieldn(n, fun, args):
    for i in range(n):
        yield fun(*args)

def foursizes():
    for dim in ((30,30), (60,60), (90,90), (120,120)):
        for x in yieldn(3, example, (dim,)):
            yield x
        

args = {"iterator": foursizes(),
        "save": [],
        "end_t": 1000}

start(**args)
