#!/usr/bin/env python

import WatorHOCA as hoca
from WatorInterface import CA_wrapper, Display, start

def example():
    (w, h) = (90, 120)

    # Second level lattice, rules, neighbourhoods
    two_nbh = (hoca.vonNeumann(),)
    two_rule = (hoca.DiffuseRule(1.1),)
    l2 = hoca.Lattice(two_nbh, two_rule, hoca.NullTMap(),
                      dim=(w/3,h/3), dtypes=("uint32","uint8"),)  

    # First level lattice, rules, neighbourhoods
    two_nbh = (hoca.vonNeumann(),)
    two_rule = (hoca.AdvWatorRule(8, 20, 20),)
    l1 = hoca.Lattice(two_nbh, two_rule, hoca.NullTMap(), max_state=2, 
                      dim=(w,h), dtypes=("uint8",)*3 + ("bool",)*2, 
                      inplace=True, on_next="self.state[3][:,:] = False; " +
                      "self.state[4][:,:] = False")

    # The map M (and M*)
    M = hoca.ModMap(3)
    hoca.connect(l1, l2, M)

    # Put it all together, and specify which parts to display
    ret = CA_wrapper(hoca.LatticeStack((l1,l2)), "HOCA", write_mode="w")
    
    ret.display.append(Display("state", 0, 0, "127*(2-data)"))
    ret.display.append(Display("state", 1, 0, "(np.max(data)-data)*255/(np.max(data)-np.min(data))"))   

    return [ret]


def yieldn(n, fun, args):
    for i in range(n):
        yield fun(*args)


args = {"iterator": yieldn(1, example, []),
        "save": [1, 2, 20, 50, 100, 400],
        "end_t": 450}

start(**args)
